#ifndef BMP_DATABASE_HPP
#define BMP_DATABASE_HPP

#include <glibmm.h>
#include <sqlite3_bmp.h>
#include <queue>

#include <boost/variant.hpp>
#include <boost/tuple/tuple.hpp>

namespace Bmp {

      namespace DB {

	  enum Exception
	  {
	      DB_EXISTS,
	      IO_ERROR,
	      SQL_EXEC_ERROR //FIXME: We need an exception type that carries the sqlite error message
	  };

	  enum ValueType
	  {
	      VALUE_TYPE_BOOL,
	      VALUE_TYPE_INT,
	      VALUE_TYPE_REAL,	//We use "REAL" here as it's sqlite3's notation, see: http://www.sqlite.org/datatype3.html
	      VALUE_TYPE_STRING  
	  }; 

	  typedef boost::variant < bool, int, double, Glib::ustring > Variant;
	  typedef std::map    < std::string,  ValueType >	    ValueMap;
	  typedef std::pair   < std::string,  ValueType >	    PairValueType;

	  typedef std::map    < std::string,  Variant >		    DataRow;
	  typedef std::vector < DataRow >			    VDataRow;
    
	  typedef std::map    < std::string,  ValueMap >	    TableMapMap;

	  /** Tuple for performing queries
            *
            * bool field denotes whether it must be an exact match or not
	    * std::string field denotes the column name
	    * Variant field holds the to-be-matched value
	    */
	  typedef boost::tuple<bool, std::string, Variant> Attribute;

	  /** Tuple for performing setting of values
            *
	    * std::string field denotes the column name
	    * Variant field holds the to-be-matched value
	    */
	  typedef boost::tuple<std::string, Variant> AttributeSet;


	  /** A vector of @link Attribute@endlink
	    *
	    */
	  typedef std::vector<Attribute> AttributeList;

	  /** A vector of @link AttributeSet@endlink
	    *
	    */
	  typedef std::vector<AttributeSet> AttributeListSet;

	  /** Slot for notifying about changes in a table
	   *
	   */
	  typedef sigc::slot < void, sqlite_int64 > SlotSQLHook;
	  typedef std::vector < SlotSQLHook > Hooks;
	  typedef std::map < std::string, Hooks > TableHookMap;

	  enum HookType
	  {
	    HOOK_INSERT,
	    HOOK_DELETE,
	    HOOK_UPDATE,

	    N_HOOKS,
	  };

	  typedef boost::tuple  < HookType, sqlite_int64, std::string > QueueHook;
	  typedef std::queue    < QueueHook >				HookQueue;
	
	  //Class "DB"
	  class DB
	  {
	    public:

	      DB (std::string name, std::string path, bool truncate = true);
	      ~DB ();

	      //Creating hooks  
	      void
	      add_hook	      (const std::string& name,
			       HookType		  type,
			       const SlotSQLHook& slot);

	      void
	      create_table    (const std::string&   name,
			       const std::string&   pkey,
			       const ValueMap&	    map);

	      void
	      create_attribute_view (const std::string&   name,
				     const std::string&   view_name,
				     const std::string&   attribute,
				     const ValueMap&	  map);	

	      void
	      create_view_manual    (const std::string&  name,
				     const std::string&  view_name,
				     const std::string&  select,
				     const ValueMap&	 map);

	      void
	      drop_table      (const std::string& name);

	      bool
	      table_exists    (const std::string& name);
  
	      void  
	      run_query	      (const Glib::ustring& query);

	      void
	      txn_begin	      ();
	      void
	      txn_commit      ();
	      void
	      txn_rollback    ();

	      void 
	      project	      (const std::string&   name,
			       const std::string&   project, 
			       VDataRow&	    vector,
			       const AttributeList& attributes = AttributeList ());

	      void 
	      get	      (const std::string&     name,
			       const AttributeList&   attributes,
			       VDataRow&	      vector,
			       const std::string&     suffix = std::string()); 

	      void 
	      get_by_rowid    (const std::string&     name,
			       sqlite_int64	      rowid,
			       DataRow&		      row); 

	      void 
	      set	      (const std::string&	name,
			       const std::string&	pkey,
			       const std::string&	pkey_value,
			       const AttributeListSet&  attributes);

	      void
	      add	      (const std::string&     name,
			       const DataRow&	      row);
			       

	      void
	      del	      (const std::string&     name,
			       const AttributeList&   attributes);

	    private:
	      static void sqlite_trace	 (void*	        data,
					  const char*   sqltext);

#if 0
	      static void sqlite_update_hook (void*	    data,
					      int	    mod,
					      char const*   database,
					      char const*   table,
					      sqlite_int64  rowid);		      
#endif

	      void
	      sqlite_exec_simple (const char* sql); 

	      TableMapMap  mapmap;
	      std::string  name;
	      std::string  path; //sqlite3 filename = path+name+".mlib";

#if 0
	      Glib::RWLock lock;
#endif
	      Glib::Mutex  lock;
	      sqlite3	   *sqlite;

#if 0
	      // Watch hooks
	      TableHookMap hooks[N_HOOKS];
#endif
	  };

      } // DB

}; // Bmp

#endif // BMP_DATABASE_HPP
