//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <boost/algorithm/string.hpp>

// Musicbrainz
#include <musicbrainz/musicbrainz.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>

#include <iostream>
#include <sstream>

#include <iostream>
#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>

// BMP
#include "uri++.hpp"
#include "database.hpp"
#include "library.hpp"
#include "file_utils.hpp"

// TagLib
#include <taglib_plugins/tplugins.h>

#include <taglib/taglib.h>
#include <taglib/fileref.h>
#include <taglib/audioproperties.h>
#include <taglib/tfile.h>
#include <taglib/tag.h>
#include <taglib/id3v2tag.h>
#include <taglib/mpegfile.h>
#include <taglib/oggfile.h>
#include <taglib/vorbisfile.h>
#include <taglib/vorbisproperties.h>
#include <taglib/xiphcomment.h>
#include <taglib/mpcfile.h>
#include <taglib/vorbisfile.h>
#include <taglib/flacfile.h>
#include <taglib/xiphcomment.h>
#include <taglib/id3v2framefactory.h>
#include <taglib/textidentificationframe.h>
#include <taglib/uniquefileidentifierframe.h>

// GStreamer
#include <gst/gst.h>
#include <gst/gstelement.h>

#include "main.hpp"
#include "paths.hpp"
#include "debug.hpp"

#include "x_library.hpp"

#ifdef HAVE_HAL
#  include "x_hal.hpp"
#endif

#define TAG_READ_TUPLE(_data) ((TagReadTuple*)(_data))

#define MLIB_VERSION "2"

namespace Bmp
{
  namespace Library
  {
    namespace
    {
      struct TagReadTuple
      {
          DB::DataRow *row;
          bool         eos;
      };

      // We're using GST-style tag identifiers here, except that we can't use dashes as that seems to confuse sqlite
      DatumDefine track_attributes[] = {
          {
              N_("Location"),
              N_("Locations"),
              N_("The location of the track (file, internet stream, etc.)"),
              "location",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Artist"),
              N_("Artists"),
              N_("The artist/peformer of the track"),
              "artist",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Album"),
              N_("Albums"),
              N_("The album of the track"),
              "album",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Title"),
              N_("Titles"),
              N_("The title of the current track"),
              "title",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Track"),
              N_("Tracks"),
              N_("The track number of the track"),
              "tracknumber",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Time"),
              N_("Times"),
              N_("The length of the track"),
              "time",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Genre"),
              N_("Genres"),
              N_("The genre (kind of music) of the track"),
              "genre",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Comment"),
              N_("Comments"),
              N_("Comments added to the track"),
              "comment",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Rating"),
              N_("Ratings"),
              N_("The rating of the track"),
              "rating",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Date"),
              N_("Dates"),
              N_("The date (year) of the track"),
              "date",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("MTIME"),
              N_("MTIMEs"),
              N_("The modification time of the track"),
              "mtime",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Bitrate"),
              N_("Bitrates"),
              N_("The bitrate of the track"),
              "bitrate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Samplerate"),
              N_("Samplerates"),
              N_("The samplerate of the track"),
              "samplerate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Play count"),
              N_("Play counts"),
              N_("The number of times this track was played"),
              "count",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("ASIN"),
              N_("ASINs"),
              N_("Amazon Standard Index Number"),
              "asin",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Volume UDI"),
              N_("Volume UDIs"),
              N_("HAL Volume UDI this track resides on"),
              "volume_udi",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Device UDI"),
              N_("Device UDIs"),
              N_("HAL Device UDI this track resides on"),
              "device_udi",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("HAL Path"),
              N_("HAL Paths"),
              N_("Path of the track relative to it's volume's root"),
              "vrp",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Album Artist Id"),
              N_("MusicBrainz Album Artist Ids"),
              N_("The MusicBrainz Release Artist ID (The ID of the Artist this release is accredited to)"),
              "musicbrainz_albumartistid",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Track ID"),
              N_("MusicBrainz Track IDs"),
              N_("The MusicBrainz Track ID"),
              "musicbrainz_trackid",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Album Id"),
              N_("MusicBrainz Album Ids"),
              N_("The MusicBrainz Release ID"),
              "musicbrainz_albumid",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Artist Id"),
              N_("MusicBrainz Artist Ids"),
              N_("The MusicBrainz Artist ID"),
              "musicbrainz_artistid",
              DB::VALUE_TYPE_STRING,
              false,
          },

      };

      TagLib::ID3v2::UserTextIdentificationFrame*
      find_utif (TagLib::ID3v2::Tag *tag, const TagLib::String &description)
      {
          TagLib::ID3v2::FrameList l = tag->frameList("TXXX");
          for(TagLib::ID3v2::FrameList::Iterator it = l.begin(); it != l.end(); ++it) {
            TagLib::ID3v2::UserTextIdentificationFrame *f = dynamic_cast<TagLib::ID3v2::UserTextIdentificationFrame *>(*it);
            if(f && f->description() == description)
              return f;
          }
          return 0;
      }

      // Based on SoundJuicer code, (C) Ross Burton <ross@burtonini.com> 
      void
      add_txxx_tag (TagLib::ID3v2::Tag *id3v2tag,
                    const Glib::ustring&   spec_id,
                    const Glib::ustring&   realworld_id,
                    const Glib::ustring&   id_str) 
      {
      using namespace Bmp;
      using namespace TagLib;

      ID3v2::UserTextIdentificationFrame * frame;

      if (id_str.empty()) return;

      if (!spec_id.empty())
      {
        frame = new ID3v2::UserTextIdentificationFrame (String::UTF8);
        id3v2tag->addFrame (frame);
        frame->setDescription (String (spec_id.c_str(), String::UTF8));
        frame->setText (String (id_str.c_str(), String::UTF8));
      }

      if (!realworld_id.empty())
      {
        frame = new ID3v2::UserTextIdentificationFrame (String::UTF8);
        id3v2tag->addFrame (frame);
        frame->setDescription (String (realworld_id.c_str(), String::UTF8));
        frame->setText (String (id_str.c_str(), String::UTF8));;
      }
      }

      int
      get_duration_from_sectors (int sectors)
      {
          const int bytes_per_sector   = 2352;
          const int bytes_per_seconds  = (44100 / 8) / 16 / 2;

          return (sectors * bytes_per_sector / bytes_per_seconds)/1000;
      }

      void
      tag_foreach (const GstTagList *list,
                   const gchar      *tag,
                   TagReadTuple     *tuple)
      {
          int count = gst_tag_list_get_tag_size (list, tag);

          for (int i = 0; i < count; i++)
            {
              const GValue *value = gst_tag_list_get_value_index (list, tag, i);

              if (G_VALUE_HOLDS (value, G_TYPE_INT))
                {
                  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_int(value))));
                }
              else
              if (G_VALUE_HOLDS (value, G_TYPE_UINT))
                {
                  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_uint(value))));
                }
              else
              if (G_VALUE_HOLDS (value, G_TYPE_STRING))
                {
                  const char *_v_C = g_value_get_string (value);
                  if (*_v_C)
                    {
                      std::string v = _v_C; 
                      boost::trim (v);
                      tuple->row->insert (std::make_pair (std::string (tag), v)); 
                    }
                }
            }
      }

      gboolean
      metadata_bus_handler (GstBus       *bus,
                            GstMessage   *message,
                            TagReadTuple *tuple)
      {
          switch (GST_MESSAGE_TYPE (message))
          {
          case GST_MESSAGE_EOS:
              {
                  tuple->eos = true;
              }
              break;

          case GST_MESSAGE_TAG:
              {
                  GstTagList *tag_list = 0;

                  gst_message_parse_tag (message, &tag_list);
                  if (tag_list)
                    {
                      gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach, tuple);
                      gst_tag_list_free (tag_list);
                    }
              }
              break;

          case GST_MESSAGE_ERROR:
              {
                  GError *error = 0;
                  gst_message_parse_error (message, &error, 0);
                  g_print ("%s ERROR: %s\n", G_STRLOC, error->message);
                  g_error_free (error);
              }
              break;

          default:
              break;
          }

          return FALSE;
      }

      void
      metadata_event_loop (TagReadTuple *tuple,
                           GstElement   *element,
                           gboolean      block)
      {
          GstBus *bus = gst_element_get_bus (element);
          g_return_if_fail (bus != 0);

          bool done = false;

          while (!done && !tuple->eos)
          {
              GstMessage *message;

              if (block)
                  message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);
              else
                  message = gst_bus_pop  (bus);

              if (message == 0)
              {
                  gst_object_unref (bus);
                  return;
              }

              done = metadata_bus_handler (bus, message, tuple);
              gst_message_unref (message);
          }
          gst_object_unref (bus);
      }

    } // anonymous namespace

    void
    get_audiocd_metadata_offline (MusicBrainz &o, DB::VDataRow *vector)
    {
        if (o.Query (std::string (MBQ_GetCDTOC)))
        {
            g_log (BMPX_SYSTEM_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: Unable to get read CD TOC", G_STRLOC);
        return;
        }

        unsigned int n_tracks = o.DataInt (MBE_TOCGetLastTrack);

        for (unsigned int i = 1; i <= n_tracks; i++)
        {
            DB::DataRow row;

            std::string uri = (boost::format ("cdda:///%d") % i).str ();
            std::string title = (boost::format ("Track %d") % i).str ();

            row.insert (std::make_pair (library->get_metadatum_id(DATUM_LOCATION), uri));
            row.insert (std::make_pair (library->get_metadatum_id(DATUM_TITLE), title));
            row.insert (std::make_pair (library->get_metadatum_id(DATUM_TRACK), int(i)));
            row.insert (std::make_pair (library->get_metadatum_id(DATUM_TIME),  int (get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i)))));

            vector->push_back (row);
        }
    }

    void
    get_audiocd_metadata (int &status, DB::VDataRow &vector)
    {
        // Create the musicbrainz object, which will be needed for subsequent calls
        MusicBrainz o;
        o.UseUTF8 (true);
        o.SetDevice (mcs->key_get<std::string>("audio", "cdrom-device"));

        bool isMultipleArtist = false;
        int trackNum = 0;

        int ret;
        ret = o.Query (MBQ_GetCDTOC);
        if (!ret)
        {
            string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
        return;
        }

        std::string cdindex = o.Data(std::string (MBE_TOCGetCDIndexId));
        if (cdindex.empty())
        {
            std::string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
        return;
        }

        ret = o.Query(std::string (MBQ_GetCDInfo));
        if (!ret)
        {
            std::string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
        return;
        }

        // Select the first album
        std::list<int> argList;
        argList.push_back (1);
        if (!o.Select (std::string (MBS_SelectAlbum), &argList))
        {
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
        return;
        }

        std::string album_id_url = o.Data(std::string (MBE_AlbumGetAlbumId));
        if (album_id_url.empty ()) return;

        std::string album_id;
        o.GetIDFromURL (album_id_url, album_id);

        // Extract the album name
        std::string album_name = o.Data(std::string (MBE_AlbumGetAlbumName));

        // Extract the number of tracks
        int n_tracks = o.DataInt(std::string (MBE_AlbumGetNumTracks));

        if (n_tracks < 1)
        {
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
        return;
        }

        // Check to see if there is more than one artist for this album
        std::string artist_check;
        for (int i = 1; i <= n_tracks; i++)
        {
            std::string artist;
            artist = o.Data (std::string (MBE_AlbumGetArtistId));
            if (artist.empty ())
                break;

            if (i == 1)
            {
                artist_check = artist;
                continue;
            }

            if (artist.compare (artist_check))
            {
                isMultipleArtist = true;
                break;
            }
        }

        std::string artist_name, artist_id;

        if (!isMultipleArtist)
        {
            // Extract the artist name from the album
            artist_name = o.Data (std::string (MBE_AlbumGetArtistName), 1);

            // Extract the artist id from the album
            artist_id   = o.Data (std::string (MBE_AlbumGetArtistId), 1);
        }

        for(int i = 1; i <= n_tracks; i++)
        {
            DB::DataRow row;
            std::stringstream uri;

            uri << "cdda://" << cdindex << "/" << i;

            row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri.str()));
            row.insert (std::make_pair (track_attributes[DATUM_ALBUM].id, Glib::ustring (album_name)));
            row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(1411)));
            row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(44100)));

            // Extract the track name from the album.
            std::string title = o.Data (std::string (MBE_AlbumGetTrackName), i);
            if (!title.empty ())
            {
                row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, DB::Variant (Glib::ustring (title))));
            }

            // Extract the album id from the track. Just use the
            // first album that this track appears on
            std::string track_id = o.Data (std::string (MBE_AlbumGetTrackId), i);
            if (!track_id.empty ())
            {
                // Extract the track number
                trackNum = o.GetOrdinalFromList (std::string (MBE_AlbumGetTrackList), track_id);
                if (trackNum > 0 && trackNum < 100)
                {
                    row.insert (std::make_pair (track_attributes[DATUM_TRACK].id,
                                                 DB::Variant (int(trackNum))));
                }
            }

            // If its a multple artist album, print out the artist for each track
            if (isMultipleArtist)
            {
                // Extract the artist name from this track
                string individual_artist;
                individual_artist = o.Data(std::string (MBE_AlbumGetArtistName), i);

                if (!individual_artist.empty ())
                {
                    row.insert (std::make_pair (track_attributes[DATUM_ARTIST].id,
                                                 DB::Variant (Glib::ustring (individual_artist))));
                }
            }
            else
            {
                row.insert (std::make_pair (track_attributes[DATUM_ARTIST].id,
                                             DB::Variant (Glib::ustring (artist_name))));
            }

            vector.push_back (row);
        }

        ret = o.Query(MBQ_GetCDTOC);
        if (!ret)
        {
            std::string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
        return;
        }

        for(int i = 1; i <= n_tracks; i++)
        {
            DB::DataRow& row = vector[i-1];
            int duration = get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i+1));
            row.insert (std::make_pair (track_attributes[DATUM_TIME].id, duration)); 
    } 

        status = Bmp::Library::Library::MB_OK;
    }

    void
    Library::metadata_set_taglib_common (const UpdateTrack& track,
                                         TagLib::Tag* tag)
    {
      using namespace TagLib;
      if (!tag) return;

      if (track.title)
          tag->setTitle   (String (track.title.get().c_str(), String::UTF8)); 

      if (track.artist)
          tag->setArtist  (String (track.artist.get().c_str(), String::UTF8));

      if (track.album)
          tag->setAlbum   (String (track.album.get().c_str(), String::UTF8));

      if (track.comment)
          tag->setComment (String (track.comment.get().c_str(), String::UTF8));

      if (track.genre)
          tag->setGenre   (String (track.genre.get().c_str(), String::UTF8));

      if (track.date)
          tag->setYear    (track.date.get());

      if (track.tracknumber)
          tag->setTrack   (track.tracknumber.get());
    }

    bool 
    Library::metadata_set_taglib (const Glib::ustring&  location,
                                  const UpdateTrack&    track)
    {
        std::string filename = Glib::filename_from_uri (location); 

        TagLib::File *file = 0;

        if (str_has_suffix_nocase (filename.c_str(), "mp3"))
          {
            using namespace TagLib;
            if (!Glib::file_test (filename, Glib::FILE_TEST_EXISTS)) return false;
            ID3v2::FrameFactory *factory = ID3v2::FrameFactory::instance(); 
            factory->setDefaultTextEncoding (String::UTF8);
            MPEG::File *file = new MPEG::File (filename.c_str(), factory);
            if (!file) return false;

            ID3v2::Tag *tag = file->ID3v2Tag (true);

            metadata_set_taglib_common (track, tag);

            if (track.mb_artistid)
            {
              add_txxx_tag (tag,
                  track_attributes[DATUM_MB_ARTIST_ID].description,
                  track_attributes[DATUM_MB_ARTIST_ID].id,
                  track.mb_artistid.get());
            }

            if (track.mb_albumid)
            {
              add_txxx_tag (tag,
                  track_attributes[DATUM_MB_ALBUM_ID].description,
                  track_attributes[DATUM_MB_ALBUM_ID].id,
                  track.mb_albumid.get());
            }

            if (track.mb_album_artistid)
            {
              add_txxx_tag (tag,
                  track_attributes[DATUM_MB_ALBUM_ARTIST_ID].description,
                  track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                  track.mb_album_artistid.get());
            }

            if (track.mb_trackid)
            {
              ID3v2::UniqueFileIdentifierFrame * frame;
              frame = new ID3v2::UniqueFileIdentifierFrame ("http://musicbrainz.org", track.mb_trackid.get().c_str());
              tag->addFrame (frame);
            }

            if (track.asin)
            {
              add_txxx_tag (tag,
                  track_attributes[DATUM_ASIN].description,
                  track_attributes[DATUM_ASIN].id,
                  track.asin.get());
            }

            file->save();
            delete file;
            return true;
          }
        else 
        if (str_has_suffix_nocase (filename.c_str(), "ogg"))
          {
            using namespace TagLib;
            if (!Glib::file_test (filename, Glib::FILE_TEST_EXISTS)) return false;
            file = new TagLib::Ogg::Vorbis::File (filename.c_str());
            if (!file) return false;

            metadata_set_taglib_common (track, file->tag());
            TagLib::Ogg::XiphComment *comment = reinterpret_cast<TagLib::Ogg::Vorbis::File*>(file)->tag (); 

            if (comment)
              {
                if (track.mb_artistid)
                {
                  comment->addField (String ("MUSICBRAINZ_ARTISTID", String::UTF8), String(track.mb_artistid.get(), String::UTF8));
                }

                if (track.mb_albumid)
                {
                  comment->addField (String ("MUSICBRAINZ_ALBUMID", String::UTF8), String(track.mb_albumid.get(), String::UTF8));
                }

                if (track.mb_album_artistid)
                {
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTISTID", String::UTF8), String(track.mb_album_artistid.get(), String::UTF8));
                }

                if (track.mb_trackid)
                {
                  comment->addField (String ("MUSICBRAINZ_TRACKID", String::UTF8), String(track.mb_trackid.get(), String::UTF8));
                }

                if (track.asin)
                {
                  comment->addField (String ("ASIN", String::UTF8), String(track.asin.get(), String::UTF8));
                }
              }
            file->save ();
            delete file;
            return true;
        }
      else
      if (str_has_suffix_nocase (filename.c_str(), "flac"))
        {
          if (!Glib::file_test (filename, Glib::FILE_TEST_EXISTS)) return false;
          file = new TagLib::FLAC::File (filename.c_str());
        }
      else
      if (str_has_suffix_nocase (filename.c_str(), "mpc"))
        {
          if (!Glib::file_test (filename, Glib::FILE_TEST_EXISTS)) return false;
          file = new TagLib::MPC::File (filename.c_str());
        }

      if (file && file->isValid() && file->tag() && !file->readOnly())
        {
          metadata_set_taglib_common (track, file->tag());
          file->save();
          delete file;
          return true;
        }
      return false;
    }

    void 
    Library::metadata_get_taglib (const   Glib::ustring &uri,
                                  DB::DataRow           &row)
    {
      using namespace TagLib;
      using namespace boost;

      std::string filename;
      try {
            filename = Glib::filename_from_uri (uri.c_str()); 
          }
      catch (Glib::ConvertError& cxe) { throw INVALID_FILE; } 

      if (!File::isReadable (filename.c_str())) throw NO_METADATA;

      File *track = 0;

      const Datum extra_tags[] =
      {
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_ASIN,
      };

      const Datum extra_tags_vorbis[] =
      {
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_TRACK_ID,
        DATUM_ASIN,
      };

      TagLib::FileRef fileref;
      fileref = TagLib::FileRef (filename.c_str());

      if (fileref.isNull()) throw NO_METADATA;
      if (!fileref.file()->isValid()) throw NO_METADATA;

      if (str_has_suffix_nocase (filename.c_str(), "mp3"))
        {
          using namespace TagLib;

          MPEG::File *file;

          if (!(file = dynamic_cast<MPEG::File*> (fileref.file()))) throw NO_METADATA;
          ID3v2::Tag *tag = file->ID3v2Tag (false);
          
          if (tag) 
            {
              using boost::algorithm::split;
              using boost::algorithm::is_any_of;

              ID3v2::UserTextIdentificationFrame *frame = 0;

              for (unsigned int n = 0; n < G_N_ELEMENTS(extra_tags); ++n) 
              {
                  frame = find_utif (tag,
                             String (track_attributes[extra_tags[n]].id, String::UTF8));
                  if (frame)
                    {
                      std::vector<std::string> subs;
                      std::string str = frame->toString().to8Bit(true);
                      split (subs, str, is_any_of (" "));
                      row.insert (std::make_pair (track_attributes[extra_tags[n]].id, subs[1])); 
                    }
              }

              // Check for UFID
              using TagLib::ID3v2::FrameList;
              FrameList const& map = tag->frameListMap()["UFID"];
              if (!map.isEmpty())
                {
                  for (FrameList::ConstIterator iter = map.begin(), end = map.end (); iter != end; ++iter)
                    {
                    ID3v2::UniqueFileIdentifierFrame *ufid = reinterpret_cast<ID3v2::UniqueFileIdentifierFrame*>(*iter);
                    if (ufid->owner() == "http://musicbrainz.org")
                      {
                        ByteVector vec = ufid->identifier();
                        vec.append ('\0');
                        row.insert (std::make_pair (track_attributes[DATUM_MB_TRACK_ID].id, Glib::ustring(vec.data())));
                        break;
                      }
                    }
                }
              }
          track = file;
        }
      else
      if (str_has_suffix_nocase (filename.c_str(), "ogg"))
        {
          using namespace TagLib;

          Ogg::Vorbis::File *file = 0;

          if (!(file = dynamic_cast<Ogg::Vorbis::File*> (fileref.file()))) throw NO_METADATA;
          Ogg::XiphComment *comment = dynamic_cast<Ogg::Vorbis::File*>(file)->tag (); 

          if (comment)
            {
              TagLib::Map<TagLib::String, TagLib::StringList> const& map = comment->fieldListMap();
              TagLib::Map<TagLib::String, TagLib::StringList>::ConstIterator iter;

              const char* mbOggIdentifiers[] =
              {
                "MUSICBRAINZ_ALBUMARTISTID",
                "MUSCIBRAINZ_ALBUMID",
                "MUSICBRAINZ_ARTISTID",
                "MUSICBRAINZ_TRACKID",
                "ASIN",
              };

              for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_vorbis); ++n)
              {
                iter = map.find (mbOggIdentifiers[n]);
                if (iter == map.end()) continue;
                TagLib::StringList const& list = iter->second;
                row.insert (std::make_pair (track_attributes[extra_tags_vorbis[n]].id, list[0].to8Bit(true)));
              }
            }
          track = file;
        }
      else
        {
          track = fileref.file ();
        }

      if (!track) throw NO_METADATA; 

      struct stat file_stat;
      stat (filename.c_str(), &file_stat);
      std::string value;

      row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id,   Glib::ustring(uri)));
      row.insert (std::make_pair (track_attributes[DATUM_TRACK].id,      int (track->tag()->track())));
      row.insert (std::make_pair (track_attributes[DATUM_DATE].id,       int (track->tag()->year())));

      value = track->tag()->artist().to8Bit (true);

      if (!value.empty())
        {
          trim (value);
          row.insert (std::make_pair (track_attributes[DATUM_ARTIST].id, value)); 
        }

      value = track->tag()->album().to8Bit (true);
      if (!value.empty())
        {
          trim (value);
          row.insert (std::make_pair (track_attributes[DATUM_ALBUM].id, value)); 
        }

      value = track->tag()->genre().to8Bit (true);
      if (!value.empty())
        {
          trim (value);
          row.insert (std::make_pair (track_attributes[DATUM_GENRE].id, value));
        }

      value = track->tag()->comment().to8Bit (true);
      if (!value.empty())
        {
          trim (value);
          row.insert (std::make_pair (track_attributes[DATUM_COMMENT].id, value)); 
        }

      row.insert (std::make_pair (track_attributes[DATUM_RATING].id,     int (0)));
      row.insert (std::make_pair (track_attributes[DATUM_COUNT].id,      int (0)));
      row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id,    int (track->audioProperties()->bitrate())));
      row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int (track->audioProperties()->sampleRate())));
      row.insert (std::make_pair (track_attributes[DATUM_TIME].id,       int (track->audioProperties()->length())));

      if  ((!track->tag()->title().toCString(true)) || (!std::strlen(track->tag()->title().toCString(true))))
        {
              row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, std::string (Glib::path_get_basename(filename))));
        }
      else
        {
          value = track->tag()->title().to8Bit(true);
          if (!value.empty())
            {
              trim (value);
              row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, value)); 
            }
        }
    }

    void 
    Library::metadata_get_gst  (const Glib::ustring&  uri,
                                DB::DataRow&          row)
    {
      std::string filename = Glib::filename_from_uri (uri);

      if (!Glib::file_test (filename, Glib::FILE_TEST_EXISTS)) return;

          boost::shared_ptr<TagReadTuple> ttuple;

          ttuple = boost::shared_ptr<TagReadTuple> (new TagReadTuple);
          ttuple->eos = false;
          ttuple->row = &row;

          ttuple->row->insert (std::make_pair (get_metadatum_id (DATUM_LOCATION), std::string (uri)));

          GstElement *element   = gst_element_factory_make ("playbin",  "playbin0");
          GstElement *fakesink  = gst_element_factory_make ("fakesink", "fakesink0");
          GstElement *videosink = gst_element_factory_make ("fakesink", "fakesink1");

          g_object_set (G_OBJECT (element), "uri", uri.c_str(), "audio-sink", fakesink, "video-sink", videosink, NULL);

          GstStateChangeReturn state_ret = gst_element_set_state (element, GST_STATE_PAUSED);
          gint                 change_timeout = 5;

          while (state_ret == GST_STATE_CHANGE_ASYNC && change_timeout > 0)
            {
              GstState state;
              state_ret = gst_element_get_state (GST_ELEMENT (element), &state, 0, 1 * GST_SECOND);
              change_timeout--;
            }

          if (state_ret != GST_STATE_CHANGE_FAILURE && state_ret != GST_STATE_CHANGE_ASYNC)
            {
              // Post application specific message so we'll know when to stop the message loop
              GstBus *bus;
              bus = gst_element_get_bus (GST_ELEMENT (element));
              if (bus)
                {
                  gst_bus_post (bus, gst_message_new_application (GST_OBJECT (element), 0));
                  gst_object_unref (bus);
                }
              // Poll the bus for messages
              metadata_event_loop (ttuple.get(), GST_ELEMENT (element), FALSE);
            }

          GstFormat  format = GST_FORMAT_TIME;
          GstQuery  *query  = gst_query_new_duration (format);

          if (gst_element_query (GST_ELEMENT(element), query))
            {
              gst_element_get_state (GST_ELEMENT(element), 0, 0, 100 * GST_MSECOND);
              gint64 length_in_nanoseconds;
              gst_query_parse_duration (query, &format, &length_in_nanoseconds);
              int length = length_in_nanoseconds / GST_SECOND;
              ttuple->row->insert (std::make_pair (get_metadatum_id (DATUM_TIME), DB::Variant (length)));
            }
          gst_query_unref (query);
          gst_element_set_state (GST_ELEMENT (element), GST_STATE_NULL);
          gst_object_unref (GST_ELEMENT (element));
    }

    Glib::ustring
    Library::get_titlestring (const Glib::ustring& uri, const std::string& format_)
    {
      GString     *string_out   = 0;
      GString     *string_field = 0;
      char    *field;
      const char  *format = format_.c_str();
      int      n;
      bool     hit = false;

      Bmp::DB::DataRow row;

      get_metadata (uri, row); 

      if (row.empty())   
      {
        throw NO_METADATA;
      }

      string_out = g_string_new ("");
      while (*format)
      {
      // Begin sequence
      if ((*format == '%') && (*(format+1) != 0) && (*(format+1) == '{'))
        {
          string_field = g_string_new ("");
          format += 2;
          while (*format && (*format != '}') && (*format != ' '))
            {
          g_string_append_c (string_field, *format);
          format++;
            }

          field = g_string_free (string_field, FALSE);
          if (!*format)
            {
              free (field);
          char *str = g_string_free (string_out, FALSE);
          std::string stdstr = str;
          free (str);
          return stdstr;
            }

          // Append metadata field
          for (n = 0; n < Bmp::Library::DATUM_LAST; ++n)
            {
              if (!g_ascii_strcasecmp (field, get_metadatum_id (Bmp::Library::Datum (n)).c_str())) 
                {
                  Bmp::DB::Variant v = row[std::string(field)];
                  char *value = 0;

                  if (Bmp::DB::ValueType(v.which()) == Bmp::DB::VALUE_TYPE_INT)
                    {
                      value = g_strdup_printf ("%d", boost::get<int>(v)); 
                    }
                  else if (Bmp::DB::ValueType(v.which()) == Bmp::DB::VALUE_TYPE_STRING)
                    {
                      value = strdup (boost::get<Glib::ustring>(v).c_str());
                    }

                  if (value)
                    {  
                      g_strstrip (value);
                      if (strlen(value) && strcmp(value, "0") != 0)
                        {
                          hit = true;
                          g_string_append (string_out, value);
                        }
                      free (value);
                    } 
                  break;
               }
            }
          free (field);
        }
      else
        {
          g_string_append_c (string_out, *format);
        }
      format++;
    }

    if (!hit)
      {
        g_string_free (string_out, TRUE);
        return Glib::path_get_basename (uri);
      }
    else
      {
        char *str = g_string_free (string_out, FALSE);
        Glib::ustring ustr = str;
        free (str);
        return ustr;
      }
    }

    void 
    Library::add_tracks (const Glib::ustring &uri)
    {
      DB::DataRow row;

      //FIXME: We need to collect a list of files that couldn't be added

      try {
             metadata_get_taglib (uri, row);
          }
      catch (Bmp::Library::Exception& cxe)
          {
             if (cxe != INVALID_FILE) metadata_get_gst (uri, row);
          }

      if (!row.empty ())
        {
          try {
                Track track (row);
                if (track.artist && track.album && track.title)
                  {
#ifdef HAVE_HAL
                    try {
                          HALVolume volume = hal->get_volume_for_uri (uri);
                          row.insert (std::make_pair (track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi));
                          row.insert (std::make_pair (track_attributes[DATUM_HAL_DEVICE_UDI].id, volume.device_udi));
                          Bmp::URI u (track.location.get());
                          Glib::ustring vrp = Glib::filename_from_uri(uri).substr (volume.mount_path.length());
                          row.insert (std::make_pair (track_attributes[DATUM_HAL_PATH].id, vrp));
                        }
                     catch (HalException& cxe) {}
#endif
                    try {
                          db->add ("main", row);
                        }
                    catch (Bmp::DB::Exception& cxe) {}
                  }
            }
          catch (...) {}
        }
    }

    void
    Library::del (const DB::AttributeList &attrs)
    {
      signal_processing_start_.emit ();
      db->del("main", attrs);
      signal_processing_end_.emit ();
    }

    void
    Library::query (const DB::AttributeList &attrs,
            DB::VDataRow        &vector)
    {
      Glib::ustring suffix;
      suffix.append ("ORDER BY ");
      suffix.append (track_attributes[DATUM_TRACK].id);
      db->get ("main", attrs, vector, suffix);
    }

    bool
    Library::is_compilation (const Glib::ustring& album)
    {
      return (compilations.find (album) != compilations.end());
    }

    void
    Library::set_asin_for_album (const Glib::ustring& artist, const Glib::ustring& album, const Glib::ustring& asin)
    {
      char *query = sqlite3_mprintf ("UPDATE main SET asin='%q' WHERE artist='%q' AND album='%q';",
                     asin.c_str(), artist.c_str(), album.c_str()); 
      db->run_query (query);
      free (query);
    }

    Glib::ustring
    Library::get_release_for_album (const Glib::ustring& album) 
    {
      using namespace Bmp::DB;

      AttributeList attrs;
      attrs.push_back (Attribute (true, track_attributes[DATUM_ALBUM].id, album));

      VDataRow vector;
      db->get ("mbreleases", attrs, vector);

      if (vector.empty ())
        {
          throw NO_RELEASE;
        }

      DataRow row = vector[0];
      return boost::get<Glib::ustring>(row[track_attributes[DATUM_MB_ALBUM_ID].id]);
    }

    Glib::ustring
    Library::get_asin_for_album (const Glib::ustring& album)    
    {
      using namespace Bmp::DB;

      AttributeList attrs;
      attrs.push_back (Attribute (true, track_attributes[DATUM_ALBUM].id, album));

      VDataRow vector;
      db->get ("asin", attrs, vector);

      if (vector.empty ())
        {
          throw NO_ASIN;
        }

      DataRow row = vector[0];
      return boost::get<Glib::ustring>(row[track_attributes[DATUM_ASIN].id]);
    }

    /** The caller must be aware that each row in this vector contains ONLY the requrested column, re-marshaling/transforming
      * the result into something more suitable, i.e. an std::vector<> of std::string would be just too much overhead
      */
    void
    Library::project (const std::string&    project,
              DB::VDataRow      &vector,
              const DB::AttributeList   &attrs)
    {
      db->project ("main", project, vector, attrs);
    }

    GHashTable*
    Library::get_hashtable (const DB::DataRow& row)
    {
      GHashTable *table;
      GValue     *value;

      table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

      for (unsigned int n = 0; n < G_N_ELEMENTS (track_attributes); n++)
      {
        DB::DataRow::const_iterator i = row.find (track_attributes[n].id);
        if (i == row.end())
          {
            g_warning (G_STRLOC ": Attribute '%s' not found in DataRow", track_attributes[n].id);
            continue;
          }

        value = g_new0 (GValue,1);
        switch (Bmp::DB::ValueType(i->second.which()))
          {
          case Bmp::DB::VALUE_TYPE_STRING:
              g_value_init (value, G_TYPE_STRING);
              g_value_set_string (value, (boost::get<Glib::ustring>(i->second)).c_str());
              break;

          case Bmp::DB::VALUE_TYPE_INT:
              g_value_init (value, G_TYPE_INT);
              g_value_set_int (value, (boost::get<int>(i->second)));
              break;

          case Bmp::DB::VALUE_TYPE_BOOL:
              g_value_init (value, G_TYPE_BOOLEAN);
              g_value_set_boolean (value, (boost::get<bool>(i->second)));
              break;

          case Bmp::DB::VALUE_TYPE_REAL:
              g_value_init (value, G_TYPE_DOUBLE);
              g_value_set_double (value, (boost::get<double>(i->second)));
              break;
        }
        g_hash_table_insert (table, g_strdup(track_attributes[n].id), value);
      }
      return table;
    }

    void
    Library::get_metadata (const Glib::ustring  &uri,
               DB::DataRow      &row)
    {
      try {
            Bmp::URI u (uri);

            if (u.get_protocol () == Bmp::URI::PROTOCOL_HTTP)
              {
                row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri)); 
                row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(0))); 
                row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(0)));
                row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, uri));
                return;
              }

            DB::AttributeList attrs;
            DB::VDataRow vector;
            attrs.push_back (DB::Attribute (true, track_attributes[DATUM_LOCATION].id, Glib::ustring(u)));
            query (attrs, vector);

            if (vector.empty())
              {
                try {
                      metadata_get_taglib (uri, row);
                    }
                catch (Bmp::Library::Exception& cxe)
                    {
                      if (cxe != INVALID_FILE) metadata_get_gst (uri, row);
                    }

                if (row.empty())
                  {
                    row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri)); 
                    row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(0))); 
                    row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(0)));
                    row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, Glib::path_get_basename (Glib::filename_from_uri(uri))));
                  }
              }
            else
              {
                row = vector[0]; 
#ifdef HAVE_HAL
                // Let's see if the file was moved by checking with HAL
                Bmp::URI u (boost::get<Glib::ustring>(row[track_attributes[DATUM_LOCATION].id]));
                std::string filename = Glib::filename_from_uri (boost::get<Glib::ustring>(row[track_attributes[DATUM_LOCATION].id]));
                if ((u.get_protocol () == Bmp::URI::PROTOCOL_FILE) && (!Glib::file_test (filename, Glib::FILE_TEST_EXISTS)))
                  {
                  std::string  mount_path = hal->get_mount_path_for_volume_udi (boost::get<Glib::ustring>(row[track_attributes[DATUM_HAL_VOLUME_UDI].id]));
                  Glib::ustring vrp = boost::get<Glib::ustring>(row[track_attributes[DATUM_HAL_PATH].id]);
                  Glib::ustring new_uri;

                  new_uri.append ("file://");
                  new_uri.append (mount_path);
                  new_uri.append (vrp);

                  if (Glib::file_test (Glib::filename_from_uri (new_uri), Glib::FILE_TEST_EXISTS))
                  {
                      char *query = sqlite3_mprintf ("UPDATE main SET location='%q' WHERE location='%q';",
                                     new_uri.c_str(),
                                     boost::get<Glib::ustring>(row[track_attributes[DATUM_LOCATION].id]).c_str());
                      db->run_query (query);
                      free (query);
                      row[track_attributes[DATUM_LOCATION].id] = new_uri; 
                  }
                  else
                  {
                      // FIXME: We should throw something here and not create a default entry
                      row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri)); 
                      row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(0))); 
                      row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(0)));
                      row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, Glib::path_get_basename (filename)));
                  }
              }
#endif
            } 
          } catch (...) { throw NO_METADATA; } //FIXME: Don't catch ... here, also see FIXME at add_tracks()
    }
    
    const std::string
    Library::get_metadatum_id (Bmp::Library::Datum metadatum) const
    {
        return std::string (track_attributes[metadatum].id);
    }

    const DatumDefine 
    Library::get_metadatum_define (Bmp::Library::Datum metadatum) const
    {
        return track_attributes[metadatum];
    }

    void
    Library::increase_play_count (const Glib::ustring& uri)
    {
    Glib::ustring sql;
    sql.append ("UPDATE main SET ").
        append(track_attributes[DATUM_COUNT].id).
        append(" = ").
        append(track_attributes[DATUM_COUNT].id).
        append(" +1 WHERE ").
        append(track_attributes[DATUM_LOCATION].id).
        append(" = '").
        append(uri).
        append("';");
    db->run_query (sql);
    }

    bool
    Library::add_audiocd (const std::string& cdindex_id)
    {
        int status;
        DB::VDataRow rows;
        get_audiocd_metadata (status, rows);
        if (status != Bmp::Library::Library::MB_OK) return false;

    DB::AttributeList attrs;
    DB::VDataRow vector;

    std::stringstream value;
    value << "cdda://" << cdindex_id << "/1";
        attrs.push_back (DB::Attribute (true, track_attributes[DATUM_LOCATION].id, value.str ()));
  
    query (attrs, vector);
    if (vector.empty())
    {
      for (DB::VDataRow::const_iterator iter = rows.begin (); iter != rows.end (); ++iter)
      {
        db->add ("main", (*iter));
      }
    }
        return true;
    }

    void
    Library::recreate_db ()
    {
    delete db;
    db = new ::Bmp::DB::DB ("library", BMP_PATH_DATA_DIR, true);
    ::Bmp::DB::ValueMap meta;
    meta.insert (std::make_pair ("version", ::Bmp::DB::VALUE_TYPE_STRING));
    db->create_table  ("meta",
               "version",
                meta);
    Glib::ustring sql;
    sql.append ("INSERT INTO meta (version) VALUES ('").append(std::string(MLIB_VERSION)).append("');");
    db->run_query (sql);
    }

    Library::Library ()
    {
      using namespace DB;
      using namespace std;
  
      register_taglib_plugins (); 

      for (unsigned int n = 0; n < G_N_ELEMENTS(track_attributes); n++)
        {
          map.insert (std::make_pair (std::string (track_attributes[n].id), track_attributes[n].type));
        }

      db = new ::Bmp::DB::DB ("library", BMP_PATH_DATA_DIR, false);
      ::Bmp::DB::ValueMap meta;
      meta.insert (std::make_pair ("version", ::Bmp::DB::VALUE_TYPE_STRING));
      db->create_table  ("meta",
                 "version",
                  meta);
      ::Bmp::DB::AttributeList attrs;
      attrs.push_back (::Bmp::DB::Attribute (true, std::string("version"), std::string(MLIB_VERSION))); 
      ::Bmp::DB::VDataRow vrows;
      try {
        db->get ("meta", attrs, vrows);
        if (vrows.empty())
        { 
          recreate_db ();
        }
      else
        {
          Glib::ustring version = boost::get<Glib::ustring>((*(vrows[0].find ("version"))).second);
          if (version != Glib::ustring (MLIB_VERSION)) recreate_db ();
        }
      } catch (Bmp::Library::Exception& cxe) { recreate_db (); }

      db->create_table  ("main",                    //table name
                 string (track_attributes[DATUM_LOCATION].id),  //pkey identifier/column name
                 map);                      //mapping for the query (must be the same mapping used to create the table)

      //Create compilations view;
      stringstream sql;
      sql << "SELECT DISTINCT " << track_attributes[DATUM_ALBUM].id 
          << ", " 
          << "count(DISTINCT " << track_attributes[DATUM_ARTIST].id << ") AS nr FROM main GROUP BY "
          << track_attributes[DATUM_ALBUM].id << " HAVING nr > 1"; 
      map_compilations.insert (make_pair (string (track_attributes[DATUM_ALBUM].id), track_attributes[DATUM_ALBUM].type));
      map_compilations.insert (make_pair (string ("nr"), VALUE_TYPE_INT));
      db->create_view_manual ("main", "compilations", sql.str(), map_compilations);

      //Create asin view;
      sql.str("");
      sql << "SELECT DISTINCT " << track_attributes[DATUM_ALBUM].id 
          << ", " << track_attributes[DATUM_ASIN].id 
          << " FROM main WHERE length(" << track_attributes[DATUM_ASIN].id << ") AND " 
          << track_attributes[DATUM_ASIN].id << " IS NOT NULL GROUP BY "
          << track_attributes[DATUM_ALBUM].id << ";";
      map_asin.insert (make_pair (string (track_attributes[DATUM_ALBUM].id), track_attributes[DATUM_ALBUM].type));
      map_asin.insert (make_pair (string (track_attributes[DATUM_ASIN].id), track_attributes[DATUM_ASIN].type));
      db->create_view_manual ("main", "asin", sql.str(), map_asin);

      //Create mbreleases view;
      sql.str("");
      sql << "SELECT DISTINCT " << track_attributes[DATUM_ALBUM].id 
          << ", " << track_attributes[DATUM_MB_ALBUM_ID].id 
          << " FROM main WHERE length(" << track_attributes[DATUM_MB_ALBUM_ID].id << ") AND " 
          << track_attributes[DATUM_MB_ALBUM_ID].id << " IS NOT NULL GROUP BY "
          << track_attributes[DATUM_ALBUM].id << ";";
      map_releases.insert (make_pair (string (track_attributes[DATUM_ALBUM].id), track_attributes[DATUM_ALBUM].type));
      map_releases.insert (make_pair (string (track_attributes[DATUM_MB_ALBUM_ID].id), track_attributes[DATUM_MB_ALBUM_ID].type));
      db->create_view_manual ("main", "mbreleases", sql.str(), map_releases);

      update_compilations ();
    }

    void
    Library::update_compilations ()
    {
      using namespace DB;
      using namespace std;

      VDataRow rows;
      db->project ("compilations", track_attributes[DATUM_ALBUM].id, rows);
      compilations.clear ();
      for (VDataRow::iterator i = rows.begin (), end = rows.end (); i != end; ++i) 
      {
          compilations.insert (boost::get<Glib::ustring>((*i)[track_attributes[DATUM_ALBUM].id]));
      }
    }

    bool
    Library::modify_tracks (const UpdateTracks& tracks)
    {
      signal_processing_start_.emit ();
      signal_modify_start_.emit (tracks.size ());

      for (UpdateTracks::const_iterator i_Tracks      = tracks.begin (),
                                        i_TracksEnd   = tracks.end ();
                                        i_Tracks     != i_TracksEnd; ++i_Tracks)
        {
          using namespace Bmp::DB;
          UpdateTrack const& track = (*i_Tracks);

          AttributeListSet attrs;

          if (track.title)
            attrs.push_back (AttributeSet (track_attributes[DATUM_TITLE].id, track.title.get().raw())); 

          if (track.artist)
            attrs.push_back (AttributeSet (track_attributes[DATUM_ARTIST].id, track.artist.get().raw())); 

          if (track.album)
            attrs.push_back (AttributeSet (track_attributes[DATUM_ALBUM].id, track.album.get().raw())); 

          if (track.genre)
            attrs.push_back (AttributeSet (track_attributes[DATUM_GENRE].id, track.genre.get().raw())); 

          if (track.comment)
            attrs.push_back (AttributeSet (track_attributes[DATUM_COMMENT].id, track.comment.get().raw())); 

          if (track.tracknumber)
            attrs.push_back (AttributeSet (track_attributes[DATUM_TRACK].id, int(track.tracknumber.get()))); 

          if (track.date)
            attrs.push_back (AttributeSet (track_attributes[DATUM_DATE].id, int(track.date.get()))); 

          if (track.rating)
            attrs.push_back (AttributeSet (track_attributes[DATUM_RATING].id, int(track.rating.get()))); 

          if (track.asin)
            attrs.push_back (AttributeSet (track_attributes[DATUM_ASIN].id, track.asin.get().raw())); 

          if (track.mb_album_artistid)
            attrs.push_back (AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id, track.mb_album_artistid.get().raw())); 

          if (track.mb_trackid)
            attrs.push_back (AttributeSet (track_attributes[DATUM_MB_TRACK_ID].id, track.mb_trackid.get().raw())); 

          if (track.mb_albumid)
            attrs.push_back (AttributeSet (track_attributes[DATUM_MB_ALBUM_ID].id, track.mb_albumid.get().raw())); 

          if (track.mb_artistid)
            attrs.push_back (AttributeSet (track_attributes[DATUM_MB_ARTIST_ID].id, track.mb_artistid.get().raw())); 
      
          Glib::ustring modify_location = track.location.get();

#ifdef HAVE_HAL
          if (track.location_new)
          {
            HALVolume volume = hal->get_volume_for_uri (track.location_new.get());

            std::string filename1 = Glib::filename_from_uri (track.location.get());
            std::string filename2 = Glib::filename_from_uri (track.location_new.get());

            Util::copy_file (filename1, filename2);
            if (Util::compare_files (filename1, filename2))
            {
              g_unlink (filename1.c_str());
              Glib::ustring vrp = filename2.substr (volume.mount_path.length());
              attrs.push_back (AttributeSet (track_attributes[DATUM_HAL_PATH].id, vrp));
              attrs.push_back (AttributeSet (track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi));
              attrs.push_back (AttributeSet (track_attributes[DATUM_HAL_DEVICE_UDI].id, volume.device_udi));
              attrs.push_back (AttributeSet (track_attributes[DATUM_LOCATION].id, Glib::ustring ("file://") + filename2)); 
              modify_location = Glib::ustring ("file://") + filename2; 
            }
          }
#endif
          try {
                db->set ("main", track_attributes[DATUM_LOCATION].id, track.location.get (), attrs);
                metadata_set_taglib (modify_location, track);
                signal_row_updated_.emit (track.location.get(), track);   
              }
          catch (...) {}
          
          signal_modify_step_.emit (); 
          while (g_main_context_pending (NULL)) g_main_context_iteration (NULL, true);
      }   

      signal_modify_end_.emit ();
      signal_processing_end_.emit ();
      return true;
    }

    Library::~Library ()
    {
        delete db;
    }

    // Generic "processing" progress #FIXME: What is this good for?  --deadchip
    Library::SignalProcessing&
    Library::signal_processing_start ()
    { 
      return signal_processing_start_;
    }

    Library::SignalProcessing&
    Library::signal_processing_end ()
    { 
      return signal_processing_end_;
    }

    Library::SignalRowUpdated&
    Library::signal_row_updated ()
    {
      return signal_row_updated_;
    }

    // Modification progress
    Library::ModifyStart&
    Library::signal_modify_start()
    { 
      return signal_modify_start_;
    }

    Library::ModifyStep&
    Library::signal_modify_step()
    { 
      return signal_modify_step_;
    }

    Library::ModifyEnd&
    Library::signal_modify_end()
    {
      return signal_modify_end_;
    }

    //class Bmp::Library::Track
    Track::~Track ()
    {}

    Track::Track (const Bmp::DB::DataRow& row)
    {
      using namespace Bmp::DB;
      DataRow::const_iterator i, end = row.end ();

      i = row.find (track_attributes[DATUM_LOCATION].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->location = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_ARTIST].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->artist = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_ALBUM].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->album = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_TITLE].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->title = boost::get<Glib::ustring> (i->second);
      }   

      i = row.find (track_attributes[DATUM_GENRE].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->genre = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_COMMENT].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->comment = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_TRACK].id);
      if (i != end)
            this->tracknumber = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_TIME].id);
      if (i != end)
            this->duration = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_RATING].id);
      if (i != end)
            this->rating = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_DATE].id);
      if (i != end)
            this->date = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_MTIME].id);
      if (i != end)
            this->mtime = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_BITRATE].id);
      if (i != end)
            this->bitrate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_SAMPLERATE].id);
      if (i != end)
            this->samplerate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_COUNT].id);
      if (i != end)
            this->samplerate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_ASIN].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->asin = boost::get<Glib::ustring> (i->second);
      }

  #ifdef HAVE_HAL
      i = row.find (track_attributes[DATUM_HAL_VOLUME_UDI].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->volume_udi = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_HAL_DEVICE_UDI].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->device_udi = boost::get<Glib::ustring> (i->second);
      }
  #endif
      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->mb_album_artistid = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_MB_TRACK_ID].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->mb_trackid = boost::get<Glib::ustring> (i->second);
      }

      i = row.find (track_attributes[DATUM_MB_ALBUM_ID].id);
      if (i != end)
      {
        if (!(boost::get<Glib::ustring>(i->second).empty()))
            this->mb_albumid = boost::get<Glib::ustring> (i->second);
      }
    }

  } // Library namespace

} // Bmp namespace
